from click.testing import CliRunner


def invoke_cli_runner(*args, **kwargs):
    """
    Helper method to invoke the CliRunner while asserting that the exit code is actually 0.
    """

    res = CliRunner().invoke(*args, **kwargs)
    assert res.exit_code == 0, f"Got non-zero exit code {res.exit_code}. Output is: {res.output}"  # noqa: S101
    return res



import logging
import shlex
import subprocess
import sys
import typing as t
from pathlib import Path

logger = logging.getLogger()


def mp(path: Path, pattern: str) -> bool:
    """
    mp -- match path.

    Evaluate if path matches pattern, i.e. if the `glob` operation on a
    path yields any results.
    """
    return len(list(path.glob(pattern))) > 0


def run_command(command: str, errors="raise") -> t.Union[bool, Exception]:
    """
    A basic convenience wrapper around `subprocess.check_call`.
    """
    logger.info(f"Running command: {command}")
    try:
        subprocess.check_call(shlex.split(command))  # noqa: S603
    except subprocess.CalledProcessError as ex:
        if errors == "return":
            return ex
        elif errors == "bool":  # noqa: RET505
            return False
        else:
            raise
    return True


def setup_logging_old(level=logging.INFO, verbose: bool = False) -> None:
    """
    Set up logging subsystem.
    """
    log_format = "%(asctime)-15s [%(name)-10s] %(levelname)-8s: %(message)s"

    if verbose:
        level = logging.DEBUG
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def is_venv():
    """
    Identify whether the runner was invoked within a Python virtual environment.

    https://stackoverflow.com/a/42580137
    """
    return hasattr(sys, "real_prefix") or (hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix)



import importlib
import logging
import sys


def setup_logging(level=logging.INFO, verbose: bool = False):
    try:
        importlib.import_module("colorlog")
        setup_logging_colorlog(level=level, verbose=verbose)
    except ImportError:
        setup_logging_standard(level=level, verbose=verbose)


def setup_logging_standard(level=logging.INFO, verbose: bool = False):
    log_format = "%(asctime)-15s [%(name)-26s] %(levelname)-8s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def setup_logging_colorlog(level=logging.INFO, verbose: bool = False):
    import colorlog
    from colorlog.escape_codes import escape_codes

    reset = escape_codes["reset"]
    log_format = f"%(asctime)-15s [%(name)-26s] %(log_color)s%(levelname)-8s:{reset} %(message)s"

    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter(log_format))

    logging.basicConfig(format=log_format, level=level, handlers=[handler])


def tweak_log_levels():
    logging.getLogger("urllib3.connectionpool").setLevel(level=logging.INFO)



import decimal
import json


class DecimalEncoder(json.JSONEncoder):
    """
    Encoder for serializing Decimal Objects to JSON.

    https://stackforgeeks.com/blog/python-json-serialize-a-decimal-object
    """

    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalEncoder, self).default(o)



import subprocess
import typing as t
from contextlib import contextmanager

import psutil


@contextmanager
def process(*args, **kwargs) -> t.Generator[subprocess.Popen, None, None]:
    """
    Wrapper around `subprocess.Popen` to also terminate child processes after exiting.

    Implementation by Pedro Cattori. Thanks!
    -- https://gist.github.com/jizhilong/6687481#gistcomment-3057122
    """
    proc = subprocess.Popen(*args, **kwargs)  # noqa: S603
    try:
        yield proc
    finally:
        try:
            children = psutil.Process(proc.pid).children(recursive=True)
        except psutil.NoSuchProcess:
            return
        for child in children:
            child.kill()
        proc.kill()



from pathlib import Path


def list_files(path: Path, pattern: str):
    """
    Enumerate all files in given directory.
    """
    files = path.glob(pattern)
    return [item.relative_to(path) for item in files]


def list_notebooks(path: Path, pattern: str = "*.ipynb"):
    """
    Enumerate all Jupyter Notebook files found in given directory.
    """
    return list_files(path, pattern)


def list_python_files(path: Path, pattern: str = "*.py"):
    """
    Enumerate all regular Python files found in given directory.
    """
    pyfiles = []
    for item in list_files(path, pattern):
        if item.name in ["conftest.py"] or item.name.startswith("test"):
            continue
        pyfiles.append(item)
    return pyfiles


def str_list(things):
    """
    Converge list to list of strings.
    """
    return list(map(str, things))



import typing as t

import pytest


def monkeypatch_pytest_notebook_treat_cell_exit_as_notebook_skip():
    """
    Patch `pytest-notebook`, in fact `nbclient.client.NotebookClient`,
    to propagate cell-level `pytest.exit()` invocations as signals
    to mark the whole notebook as skipped.

    In order not to be too intrusive, the feature only skips notebooks
    when being explicitly instructed, by adding `[skip-notebook]` at the
    end of the `reason` string. Example:

        import pytest
        if "ACME_API_KEY" not in os.environ:
            pytest.exit("ACME_API_KEY not given [skip-notebook]")

    https://github.com/chrisjsewell/pytest-notebook/issues/43
    """
    from nbclient.client import NotebookClient
    from nbclient.exceptions import CellExecutionError
    from nbformat import NotebookNode

    async_execute_cell_dist = NotebookClient.async_execute_cell

    async def async_execute_cell(
        self,
        cell: NotebookNode,
        cell_index: int,
        execution_count: t.Optional[int] = None,
        store_history: bool = True,
    ) -> NotebookNode:
        try:
            return await async_execute_cell_dist(
                self,
                cell,
                cell_index,
                execution_count=execution_count,
                store_history=store_history,
            )
        except CellExecutionError as ex:
            if ex.ename == "Exit" and ex.evalue.endswith("[skip-notebook]"):
                raise pytest.skip(ex.evalue) from ex
            else:  # noqa: RET506
                raise

    NotebookClient.async_execute_cell = async_execute_cell  # type: ignore[method-assign,unused-ignore]



"""
-----
About
-----

Next Generation Runner (ngr): Effortless invoke programs and test harnesses.

-------
Backlog
-------
- After a few iterations, refactor to separate package.
- Look at https://pypi.org/project/ur/
"""

import logging
import os
import typing as t
from pathlib import Path

from pueblo.ngr.model import ItemType
from pueblo.ngr.runner import (
    DotNetRunner,
    ElixirRunner,
    GolangRunner,
    HaskellRunner,
    JavaRunner,
    JavaScriptRunner,
    JuliaRunner,
    MakeRunner,
    MeltanoRunner,
    PhpRunner,
    PythonRunner,
    RubyRunner,
    RunnerBase,
    RustRunner,
)

logger = logging.getLogger()


class NextGenerationRunner:
    """
    Run a <thing> based on its shape.
    """

    def __init__(self, thing: t.Any, options: t.Dict) -> None:
        self.thing = thing
        self.options = options
        self.path = Path(self.thing)
        self.runner: t.Optional[RunnerBase] = None
        self.type: t.Optional[ItemType] = None
        self.runners = {
            # Language runners, listed alphabetically.
            ItemType.DOTNET: DotNetRunner,
            ItemType.ELIXIR: ElixirRunner,
            ItemType.GOLANG: GolangRunner,
            ItemType.HASKELL: HaskellRunner,
            ItemType.JAVA: JavaRunner,
            ItemType.JAVASCRIPT: JavaScriptRunner,
            ItemType.JULIA: JuliaRunner,
            ItemType.MELTANO: MeltanoRunner,
            ItemType.PHP: PhpRunner,
            ItemType.PYTHON: PythonRunner,
            ItemType.RUBY: RubyRunner,
            ItemType.RUST: RustRunner,
            # Makefile runner, as fallback.
            ItemType.MAKE: MakeRunner,
        }
        self.identify()

    @property
    def runner_names(self):
        names = []
        for runner in self.runners.keys():
            names.append(runner.name)
        return names

    def identify(self) -> None:
        """
        Identify type of <thing>.
        """
        for type_, class_ in self.runners.items():
            logger.info(f"Probing: type={type_}, class={class_}")
            try:
                self.runner = class_(path=self.path, options=self.options)  # type: ignore[abstract]
                self.type = self.runner.type
            except Exception:
                logger.exception("Unknown error")
            if self.type is not None:
                break

        if self.type is None:
            raise NotImplementedError(
                f"Unable to identify invocation target. "
                f"Supported types: {self.runner_names}. "
                f"Target path: {self.path} "
                f"Current path: {os.getcwd()}"
            )

    def run(self) -> None:
        """
        Invoke / run <thing>.
        """
        if self.path.exists():
            if self.path.is_file():
                raise NotImplementedError(f"Invoking a file is not implemented yet ;]: {self.path}")
            elif self.path.is_dir():  # noqa: RET506
                type_name = self.type and self.type.name or None
                logger.info(f"Invoking {type_name} in directory: {self.path}")
                if self.runner is not None:
                    self.runner.run()
                else:
                    raise RuntimeError("No runner found")
            else:
                raise ValueError(f"Path is neither file nor directory: {self.path}")


def run(thing: t.Any, options: t.Dict) -> None:
    """
    Invoke the runner with a single argument, the <thing>.
    """
    ngr = NextGenerationRunner(thing=thing, options=options)
    return ngr.run()



import argparse
import logging
import sys
from argparse import ArgumentDefaultsHelpFormatter

from pueblo import __version__, setup_logging
from pueblo.ngr.core import run

logger = logging.getLogger()


def read_command_line_arguments(args=None, prog_name=None):
    """
    Parse and return command line arguments.
    """
    parser = argparse.ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter, prog=prog_name)
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="print package version of pueblo/ngr",
        version=f"%(prog)s ({__version__})",
    )

    subparsers = parser.add_subparsers(dest="command")

    test = subparsers.add_parser("test", help="Invoke test suites")
    test.add_argument("target")
    test.add_argument("--accept-no-venv", action="store_true", help="Whether to accept not running in venv")
    test.add_argument("--dotnet-version", type=str, help=".NET version, like `net6.0`, `net7.0`, or `5.0.x`, `6.0.x`")
    test.add_argument("--npgsql-version", type=str, help="Version of Npgsql")

    return parser, parser.parse_args(args=args)


def main(args=None, prog_name=None):
    """
    Main program.

    - Setup logging.
    - Read command-line parameters.
    - Run sanity checks.
    - Invoke runner.
    """
    setup_logging()
    parser, args = read_command_line_arguments(args=args, prog_name=prog_name)
    if not args.command:
        logger.error("Unable to invoke subcommand: Not given or empty")
        parser.print_help()
        sys.exit(1)
    if not args.target:
        logger.error("Unable to invoke target: Not given or empty")
        parser.print_help()
        sys.exit(1)

    try:
        run(args.target, args.__dict__)
    except NotImplementedError as ex:
        logger.critical(ex)
        sys.exit(1)



import getpass
import os
import typing as t

from dotenv import find_dotenv, load_dotenv


def init_dotenv():
    """
    Find `.env` file and load environment variables.
    """
    load_dotenv(find_dotenv())


def getenvpass(env_var: str, prompt: t.Union[str, None] = None, skip_pytest_notebook: bool = True) -> None:
    """
    Read variable from environment or `.env` file.
    If it is not defined, prompt interactively.

    FIXME: Needs a patch to make it work with `pytest-notebook`,
           see https://github.com/chrisjsewell/pytest-notebook/issues/43.
    """

    init_dotenv()
    if env_var not in os.environ:
        if "PYTEST_CURRENT_TEST" in os.environ and skip_pytest_notebook:
            import pytest

            pytest.exit(f"{env_var} not given [skip-notebook]")
        elif prompt:
            os.environ[env_var] = getpass.getpass(prompt)



from enum import Enum


class ItemType(Enum):
    DOTNET = "dotnet"
    ELIXIR = "elixir"
    GOLANG = "golang"
    HASKELL = "haskell"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    JULIA = "julia"
    MAKE = "make"
    MELTANO = "meltano"
    PHP = "php"
    PYTHON = "python"
    RUBY = "ruby"
    RUST = "rust"



import logging

import click
from click_aliases import ClickAliasedGroup

from pueblo.util.cli import boot_click

logger = logging.getLogger(__name__)


def help_pueblo():
    """
    Pueblo - a Python toolbox library.

    pueblo --version
    """  # noqa: E501


@click.group(cls=ClickAliasedGroup)
@click.version_option(package_name="pueblo")
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    return boot_click(ctx, verbose, debug)



import logging
import shlex
import shutil
import subprocess
import sys
import typing as t
from abc import abstractmethod
from pathlib import Path

from pueblo.ngr.model import ItemType
from pueblo.ngr.util import is_venv, mp, run_command

try:
    from contextlib import chdir as chdir_ctx  # type: ignore[attr-defined,unused-ignore]
except ImportError:
    from contextlib_chdir import chdir as chdir_ctx  # type: ignore[no-redef,unused-ignore]


logger = logging.getLogger()


class RunnerBase:
    def __init__(self, path: Path, options: t.Dict) -> None:
        self.path = path
        self.options = options
        self.type: t.Optional[ItemType] = None
        if hasattr(self, "__post_init__"):
            self.__post_init__()
        self.has_makefile = mp(self.path, "Makefile")
        self.peek()

    def run(self) -> None:
        # Change working directory to designated path.
        # From there, address paths relatively.

        with chdir_ctx(self.path):
            self.path = Path(".")

            logger.info("Environment Information")
            self.info()
            logger.info("Installing")
            self.install()
            logger.info("Testing")
            self.test()

    @abstractmethod
    def peek(self) -> None:
        raise NotImplementedError("Method must be implemented")

    @abstractmethod
    def info(self) -> None:
        pass

    @abstractmethod
    def install(self) -> None:
        raise NotImplementedError("Method must be implemented")

    @abstractmethod
    def test(self) -> None:
        raise NotImplementedError("Method must be implemented")


class DotNetRunner(RunnerBase):
    """
    .NET test suite runner.

    - Knows how to invoke `dotnet {restore,list,test}` appropriately.
    - Optionally accepts the `--npgsql-version=` command-line option, to specify the Npgsql version.
    """

    def __post_init__(self) -> None:
        self.has_csproj: t.Optional[bool] = None
        self.framework: t.Optional[str] = None

    def peek(self) -> None:
        self.has_csproj = mp(self.path, "*.csproj")

        dotnet_version_major = self.get_dotnet_framework_version()
        self.framework = f"net{dotnet_version_major}.0"

        if self.has_csproj:
            self.type = ItemType.DOTNET

    def get_dotnet_framework_version(self) -> int:
        """
        In:  Either one of `5.0.x`, `6.0.x`, or `net6.0`, `net7.0`,
             or fallback to `dotnet --version` => `7.0.202`.
        Out: `net5.0`, `net6.0`, `net7.0`, etc.
        """
        dotnet_version = self.options.get("dotnet_version")
        if dotnet_version:
            if dotnet_version.startswith("net"):
                return dotnet_version[3]
            dotnet_version_full = dotnet_version
        else:
            dotnet_version_full = (
                subprocess.check_output(shlex.split("dotnet --version")).decode().strip()  # noqa: S603
            )
        return int(dotnet_version_full[0])

    def info(self) -> None:
        """
        Display environment information.
        """
        run_command("dotnet --version")
        run_command("dotnet --info")

    def install(self) -> None:
        """
        Install dependencies from `.csproj` file.
        """
        self.adjust_npgql_version()
        run_command("dotnet restore")
        run_command("dotnet list . package")

    def adjust_npgql_version(self):
        npgsql_version = self.options.get("npgsql_version")
        if not npgsql_version:
            logger.info("[MATRIX] Not modifying Npgsql version")
            return
        logger.info(f"[MATRIX] Modifying Npgsql version: {npgsql_version}")
        cmd = f"""
            sed -E 's!<PackageReference Include="Npgsql" Version=".+?" />!<PackageReference Include="Npgsql" Version="{npgsql_version}" />!' \
            -i demo.csproj
        """  # noqa: E501
        run_command(cmd)

    def test(self) -> None:
        """
        Invoke `dotnet test`, with code coverage tracking.
        """
        run_command(f'dotnet test --framework={self.framework} --collect:"XPlat Code Coverage"')


class ElixirRunner(RunnerBase):
    """
    Basic Elixir runner.

    Currently, just knows to invoke `hex` and `mix` within a directory.
    """

    def __post_init__(self) -> None:
        self.mix_project: t.Optional[bool] = None

    def peek(self) -> None:
        self.mix_project = mp(self.path, "mix.exs")

        if self.mix_project:
            self.type = ItemType.ELIXIR

    def install(self) -> None:
        """
        Install package manager, and project dependencies.
        """
        run_command("mix local.hex --force --if-missing")
        run_command("mix deps.get")

    def test(self) -> None:
        """
        Run test suite.
        """
        run_command("mix test")


class GolangRunner(RunnerBase):
    """
    Basic Golang runner.

    Currently, just knows to invoke `go {build,test}` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_go_mod: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_go_mod = mp(self.path, "go.mod")

        if self.has_go_mod:
            self.type = ItemType.GOLANG

    def install(self) -> None:
        """
        Invoke `go build`.
        """
        run_command("go build")

    def test(self) -> None:
        """
        Invoke `go test -v`.
        """
        run_command("go test -v")


class HaskellRunner(RunnerBase):
    """
    Basic Haskell runner.

    Currently, just knows to invoke `stack {build,test}` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_cabal_file: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_cabal_file = mp(self.path, "*.cabal")

        if self.has_cabal_file:
            self.type = ItemType.HASKELL

    def install(self) -> None:
        """
        Install project dependencies.
        """
        run_command("stack build")

    def test(self) -> None:
        """
        Run test suite.
        """
        run_command("stack test")


class JavaRunner(RunnerBase):
    """
    Java test suite runner.

    - Knows how to invoke either Gradle or Maven.
    """

    def __post_init__(self) -> None:
        self.has_pom_xml: t.Optional[bool] = None
        self.has_gradle_files: t.Optional[bool] = None
        self.is_gradle_installed: t.Optional[bool] = None
        self.has_gradle_wrapper: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_pom_xml = mp(self.path, "pom.xml")
        self.has_gradle_files = mp(self.path, "*.gradle")
        self.is_gradle_installed = bool(shutil.which("gradle"))
        self.has_gradle_wrapper = (self.path / "gradlew").exists()

        if self.has_pom_xml or self.has_gradle_files:
            self.type = ItemType.JAVA

    def info(self) -> None:
        """
        Display environment information.
        """
        run_command("java -version")

    def install(self) -> None:
        """
        Install dependencies.
        """
        if self.has_pom_xml:
            run_command("mvn install")
        elif self.has_gradle_files:
            # When `gradle` is installed, wipe the
            # existing wrapper, and generate a new one.
            if self.is_gradle_installed:
                if self.has_gradle_wrapper:
                    (self.path / "gradlew").unlink(missing_ok=True)
                run_command("gradle wrapper")
            self.has_gradle_wrapper = (self.path / "gradlew").exists()
            if not self.has_gradle_wrapper:
                logger.error("No Gradle wrapper found, and no one could be generated. Consider installing `gradle`.")
        else:
            raise NotImplementedError("Unable to invoke target: install")

    def test(self) -> None:
        """
        Invoke software tests.
        """
        if self.has_makefile:
            run_command("make test")
        elif self.has_pom_xml:
            run_command("mvn test")
        elif self.has_gradle_files:
            run_command("./gradlew check --info")
        else:
            raise NotImplementedError("Unable to invoke target: test")


class JavaScriptRunner(RunnerBase):
    """
    Basic runner for JavaScript projects.

    Currently, just knows to invoke `npm {install,test}` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_package_json: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_package_json = mp(self.path, "package.json")

        if self.has_package_json:
            self.type = ItemType.JAVASCRIPT

    def install(self) -> None:
        """
        Invoke `npm install`.
        """
        run_command("npm install")

    def test(self) -> None:
        """
        Invoke `npm test`.
        """
        run_command("npm test")


class JuliaRunner(RunnerBase):
    """
    Basic runner for Julia projects.

    Currently, just knows to invoke `Pkg.{build,test}` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_project_toml: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_project_toml = mp(self.path, "Project.toml")

        if self.has_project_toml:
            self.type = ItemType.JULIA

    def install(self) -> None:
        """
        Invoke `Pkg.build()`.
        """
        run_command("julia --project=. --depwarn=error --eval='using Pkg; Pkg.build()'")

    def test(self) -> None:
        """
        Invoke `Pkg.test()`.
        """
        run_command("julia --project=. --depwarn=error --eval='using Pkg; Pkg.test()'")


class MakeRunner(RunnerBase):
    """
    Basic Make runner.

    Currently, just knows to invoke `make test` within a directory.
    """

    def peek(self) -> None:
        if self.has_makefile:
            self.type = ItemType.MAKE

    def install(self) -> None:
        """
        Effectively an optional noop, but let's try.
        """
        try:
            run_command("make install")
        except Exception:
            logger.exception("Command failed: make install")

    def test(self) -> None:
        """
        Invoke `make test`.
        """
        run_command("make test")


class MeltanoRunner(RunnerBase):
    """
    Invokes the `test` job on a Meltano project.
    """

    def __post_init__(self) -> None:
        self.has_requirements_txt: t.Optional[bool] = None
        self.has_meltano_yaml: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_requirements_txt = mp(self.path, "requirements*.txt")
        self.has_meltano_yaml = mp(self.path, "meltano.yml")

        if self.has_meltano_yaml:
            self.type = ItemType.MELTANO

    def install(self) -> None:
        """
        Install requirements, and run `meltano install`.
        """
        if self.has_requirements_txt:
            PythonRunner.install_requirements(self.path)
        run_command("meltano install")

    def test(self) -> None:
        """
        Invoke `meltano run test`.
        """
        run_command("meltano run test")


class PhpRunner(RunnerBase):
    """
    Basic PHP runner.

    Currently, just knows to invoke `composer` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_composer_json: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_composer_json = mp(self.path, "composer.json")

        if self.has_composer_json:
            self.type = ItemType.PHP

    def install(self) -> None:
        """
        Install dependencies of PHP Composer package.
        """
        run_command("composer install")

    def test(self) -> None:
        """
        Invoke a script called `test`, defined in `composer.json`.
        """
        run_command("composer run test")


class PythonRunner(RunnerBase):
    """
    Basic Python runner.

    Currently, just knows to invoke `pytest` within a directory.

    # TODO: Need to run within a virtualenv to be isolated from the parent environment.
    """

    def __post_init__(self) -> None:
        self.has_python_files: t.Optional[bool] = None
        self.has_setup_py: t.Optional[bool] = None
        self.has_setup_cfg: t.Optional[bool] = None
        self.has_pyproject_toml: t.Optional[bool] = None
        # TODO: Rename to `is_poetry`, by introspecting `pyproject.toml` for `[tool.poetry]`.
        self.has_poetry_lock: t.Optional[bool] = None
        self.has_requirements_txt: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_python_files = mp(self.path, "*.py")
        self.has_setup_py = mp(self.path, "setup.py")
        self.has_setup_cfg = mp(self.path, "setup.cfg")
        self.has_pyproject_toml = mp(self.path, "pyproject.toml")
        self.has_poetry_lock = mp(self.path, "poetry.lock")
        self.has_requirements_txt = mp(self.path, "requirements*.txt")
        self.is_pipx_installed = shutil.which("pipx")
        self.is_poetry_installed = shutil.which("poetry")

        if self.has_python_files or self.has_setup_py or self.has_pyproject_toml or self.has_requirements_txt:
            self.type = ItemType.PYTHON

    def run(self) -> None:
        # Sanity check. When invoking a Python thing within a sandbox,
        # don't install system-wide.
        if not is_venv():
            if not self.options.get("accept_no_venv", False):
                logger.error("Unable invoke target without virtualenv. Use `--accept-no-venv` to override.")
                sys.exit(1)

        return super().run()

    def install(self) -> None:
        """
        Install dependencies of Python thing, based on its shape.

        TODO: Heuristically figure out which "extra" packages to install from the outside.
              Example names for minimal test/development dependencies: dev,devel,tests,testing.
        """
        self.install_requirements(self.path)

        if self.has_poetry_lock:
            # TODO: Add list of extras.
            # TODO: Poetry also knows dependency groups, e.g. `--with=test`.
            if not self.is_poetry_installed:
                if not self.is_pipx_installed:
                    run_command("pip install pipx")
                run_command("pipx install poetry")
            try:
                run_command("poetry install --with=test")
            except Exception:
                run_command("poetry install")

        elif self.has_pyproject_toml or self.has_setup_cfg or self.has_setup_py:
            # By default, assume to run `pip install`.
            pip_install = True

            # `pyproject.toml` files are sometimes only used for configuring tools,
            # but not for defining project metadata. When still trying to install,
            # they will error out like `error: Multiple top-level modules discovered
            # in a flat-layout`.
            if self.has_pyproject_toml:
                if not self.pyproject_has_string("[project]"):
                    pip_install = False

            if pip_install:
                run_command("pip install --editable='.[develop,test]'")

    @staticmethod
    def install_requirements(path: Path):
        requirements_txt = list(path.glob("requirements*.txt"))
        if requirements_txt:
            pip_requirements_args = [f"-r {item}" for item in requirements_txt]
            pip_cmd = f"pip install {' '.join(pip_requirements_args)}"
            run_command(pip_cmd)

    def test(self) -> None:
        """
        Test a Python thing, based on which test runner is installed.

        First, try a high-level target of some sort of management tool,
        discover its targets, and invoke `check` or `test`.

        When no high-level tool can be discovered, just invoke `pytest`.
        """

        # On Poetry projects, invoke poetry/pytest.
        # TODO: Also allow using Poetry with other task runners, see below.
        if self.has_poetry_lock:
            # TODO: poetry run which pytest
            run_command("poetry run pytest --config-file=.")

        else:
            uses_poe = self.has_pyproject_toml and self.pyproject_has_string("[tool.poe.tasks]")
            has_pytest = shutil.which("pytest") is not None

            # When a `pyproject.toml` file includes `poethepoet` task definitions, try to
            # invoke sensible targets like `check` or `test`. This is popular amongst users
            # of `pueblo` and friends.
            # `poe check` bundles linter/checkstyle and software test targets into a single
            # entrypoint, similar to how `gradle check` is doing it.
            if uses_poe:
                candidates = ["check", "test"]
                poe_tasks = self.get_poe_tasks()
                success = False
                for candidate in candidates:
                    if candidate in poe_tasks:
                        run_command(f"poe {candidate}")
                        success = True
                        break
                if not success:
                    raise RuntimeError(f"Failed to discover poe task from candidates: {candidates}")

            elif has_pytest:
                run_command("pytest")
            else:
                raise NotImplementedError("No handler to invoke Python item")

    def pyproject_has_string(self, needle: str) -> bool:
        """
        Check whether `pyproject.toml` file contains given string.
        """
        pyproject_toml_file = self.path / "pyproject.toml"
        pyproject_toml = pyproject_toml_file.read_text()
        return needle in pyproject_toml

    def get_poe_tasks(self) -> t.List[str]:
        """
        From `poethepoet._list_tasks`.
        """
        try:
            from poethepoet.config import PoeConfig

            config = PoeConfig()
            config.load()
            return [task for task in config.tasks.keys() if task and task[0] != "_"]
        except Exception as ex:  # pylint: disable=broad-except
            # this happens if there's no pyproject.toml present
            raise ValueError(f"Discovering poe task names failed: {ex}") from ex


class RubyRunner(RunnerBase):
    """
    Basic Ruby runner.

    Currently, just knows to invoke Bundler and Rake within a directory.

    - https://bundler.io/
    - https://en.wikipedia.org/wiki/Rake_(software)
    """

    def __post_init__(self) -> None:
        self.has_gemfile: t.Optional[bool] = None
        self.has_rakefile: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_gemfile = mp(self.path, "Gemfile")
        self.has_rakefile = mp(self.path, "Rakefile")

        if self.has_gemfile or self.has_rakefile:
            self.type = ItemType.RUBY

    def install(self) -> None:
        """
        Install dependencies using Ruby's Bundler, defined in `Gemfile`.
        """
        run_command("bundle install")

    def test(self) -> None:
        """
        Invoke a rake target called `test`, defined in `Rakefile`.
        """
        run_command("bundle exec rake test")


class RustRunner(RunnerBase):
    """
    Basic Rust runner.

    Currently, just knows to invoke `cargo {build,run}` within a directory.
    """

    def __post_init__(self) -> None:
        self.has_cargo_toml: t.Optional[bool] = None

    def peek(self) -> None:
        self.has_cargo_toml = mp(self.path, "Cargo.toml")

        if self.has_cargo_toml:
            self.type = ItemType.RUST

    def install(self) -> None:
        """
        Invoke `cargo build`.
        """
        run_command("cargo build")

    def test(self) -> None:
        """
        Invoke `cargo test`.
        """
        run_command("cargo test")



import typing as t

import numpy as np
import pandas as pd


def makeTimeDataFrame(nper=None, freq="B") -> pd.DataFrame:
    """
    pandas 2.2.0 removed `pandas._testing.makeTimeDataFrame`.

    TST/CLN: Remove makeTime methods
    https://github.com/pandas-dev/pandas/pull/56264

    :param nper:
    :param freq:
    :return:
    """
    return pd.DataFrame(
        np.random.default_rng(2).standard_normal((nper, 4)),
        columns=pd.Index(list("ABCD"), dtype=object),
        index=pd.date_range("2000-01-01", periods=nper, freq=freq),
    )


def getMixedTypeDict() -> t.Tuple[pd.Index, t.Dict[str, t.Any]]:
    index = pd.Index(["a", "b", "c", "d", "e"])

    data = {
        "A": [0.0, 1.0, 2.0, 3.0, 4.0],
        "B": [0.0, 1.0, 0.0, 1.0, 0.0],
        "C": ["foo1", "foo2", "foo3", "foo4", "foo5"],
        "D": pd.bdate_range("1/1/2009", periods=5),
    }

    return index, data


def makeMixedDataFrame() -> pd.DataFrame:
    return pd.DataFrame(getMixedTypeDict()[1])



import pytest


@pytest.fixture(scope="session", autouse=True)
def nltk_init():
    """
    Initialize nltk upfront, so that it does not run stray output into Jupyter Notebooks.
    """
    download_items = ["averaged_perceptron_tagger", "punkt"]
    import nltk

    for item in download_items:
        nltk.download(item)



import contextlib
import io
import typing as t
from pathlib import Path

from pathlibfs import Path as PathPlus
from yarl import URL


@contextlib.contextmanager
def to_io(source: t.Union[str, Path, t.IO], mode: t.Literal["r", "rb", "rt"] = "rt") -> t.Generator[t.IO, None, None]:
    """
    Converge filesystem path, remote URL, or file-like object into an IO handle.
    """
    fp: t.IO
    if isinstance(source, io.TextIOWrapper):
        fp = source
    elif isinstance(source, (str, Path, PathPlus)):
        source = str(source)
        path = open_url(source)
        fp = path.open(mode=mode)
    else:
        raise TypeError(f"Unable to converge to IO handle. type={type(source)}, value={source}")
    yield fp
    fp.close()


def open_url(url: str) -> PathPlus:
    """
    Access URL, with specific handling for GitHub URLs.

    When approached using a GitHub HTTP URL, converge it to a pathlibfs / fsspec URL,
    and open it.

    Input URLs
    ----------
    github+https://foobar:ghp_lalala@github.com/acme/sweet-camino/path/to/document.md
    github+https://foobar:ghp_lalala@github.com/acme/sweet-camino/blob/main/path/to/document.md

    Output Path
    -----------
    fs = Path("github://path/to/document.md", username="foobar", token="ghp_lalala", org="acme", repo="sweet-camino")
    """
    uri = None
    try:
        uri = URL(url)
    except ValueError as ex:
        if "host is required for absolute urls" in str(ex):
            pass
        else:
            raise

    if uri and uri.scheme.startswith("github+https"):
        path_fragments = uri.path.split("/")[1:]
        path_kwargs = {
            "username": uri.user,
            "token": uri.password,
            "org": path_fragments[0],
            "repo": path_fragments[1],
        }

        real_path_fragments = path_fragments[2:]
        if path_fragments[2] in ["blob", "raw"]:
            real_path_fragments = path_fragments[4:]

        downstream_url = "github://" + "/".join(real_path_fragments)
        path = PathPlus(downstream_url, **path_kwargs)

    else:
        kwargs = {}
        if url.startswith("s3://"):
            kwargs["anon"] = True
        path = PathPlus(url, **kwargs)
    return path


def path_without_scheme(url_like: str) -> PathPlus:
    """
    Return a pathlibfs Path, without the scheme.
    """
    url = URL(str(url_like))
    if url.is_absolute():
        url = url.with_scheme("")
    return PathPlus(str(url))



import logging
import textwrap
import typing as t

from pueblo.util.logging import setup_logging

logger = logging.getLogger(__name__)


def boot_click(
    ctx: "click.Context", verbose: bool = False, debug: bool = False  # type: ignore[name-defined]  # noqa: F821
):
    """
    Bootstrap the CLI application.
    """

    # Adjust log level according to `verbose` / `debug` flags.
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG

    # Setup logging, according to `verbose` / `debug` flags.
    setup_logging(level=log_level, verbose=verbose)


def docstring_format_verbatim(text: t.Optional[str]) -> str:
    """
    Format docstring to be displayed verbatim as a help text by Click.

    - https://click.palletsprojects.com/en/8.1.x/documentation/#preventing-rewrapping
    - https://github.com/pallets/click/issues/56
    """
    text = text or ""
    text = textwrap.dedent(text)
    lines = [line if line.strip() else "\b" for line in text.splitlines()]
    return "\n".join(lines)


def make_command(cli, name, helpfun, aliases=None):
    """
    Convenience shortcut for creating a subcommand.
    """
    return cli.command(
        name,
        help=docstring_format_verbatim(helpfun.__doc__),
        context_settings={"max_content_width": 100},
        aliases=aliases,
    )



import typing as t
from pathlib import Path


def relative_to(items: t.List[Path], path: Path) -> t.List[Path]:
    """
    Compute relative path to given directory.
    """
    return [item.relative_to(path) for item in items]


def list_files(path: Path, pattern: str) -> t.List[Path]:
    """
    Enumerate all files in given directory.
    """
    return list(path.glob(pattern))


def list_directories(path: Path, pattern: str) -> t.List[Path]:
    """
    Enumerate all directories in given directory.
    """
    items = list_files(path, pattern)
    return [item for item in items if item.is_dir()]



import logging
import typing as t

import requests_cache

from pueblo.context import pueblo_cache_path

if t.TYPE_CHECKING:
    from langchain.schema import Document

http_cache_file = pueblo_cache_path / ".httpcache.sqlite"
http = requests_cache.CachedSession(str(http_cache_file))

logger = logging.getLogger(__name__)


class CachedWebResource:
    """
    A basic wrapper around `requests-cache` and `langchain`.
    """

    def __init__(self, url: str):
        logger.info(f"Using web cache: {http_cache_file}")
        self.url = url

    def fetch_single(self) -> t.List["Document"]:
        return [self.document_from_url()]

    @staticmethod
    def fetch_multi(urls) -> t.List["Document"]:
        from langchain.document_loaders import UnstructuredURLLoader

        loader = UnstructuredURLLoader(urls=urls)
        return loader.load()

    def document_from_url(self) -> "Document":
        """
        Converge URL resource into LangChain Document.
        """
        logger.info(f"Acquiring web resource: {self.url}")
        from langchain.schema import Document
        from unstructured.partition.html import partition_html

        response = http.get(self.url)
        elements = partition_html(text=response.text)
        text = "\n\n".join([str(el) for el in elements])
        metadata = {"source": self.url}
        return Document(page_content=text, metadata=metadata)

    def langchain_documents(self, **kwargs) -> t.List["Document"]:
        """
        Load URL resource, and split paragraphs in response into individual documents.
        """
        from langchain.text_splitter import CharacterTextSplitter

        documents = self.fetch_single()
        text_splitter = CharacterTextSplitter(**kwargs)
        return text_splitter.split_documents(documents)


if __name__ == "__main__":
    from pueblo import setup_logging

    setup_logging()
    url = "https://github.com/langchain-ai/langchain/raw/v0.0.325/docs/docs/modules/state_of_the_union.txt"
    docs = CachedWebResource(url).langchain_documents(chunk_size=1000, chunk_overlap=0)
    print("docs:", docs)  # noqa: T201



import importlib
import typing as t
from pathlib import Path

import pytest


def fourty_two() -> int:
    """
    A dummy function to be patched by testing machinery.
    """
    return 42


def pytest_module_function(request: pytest.FixtureRequest, filepath: t.Union[str, Path], entrypoint: str = "main"):
    """
    From individual Python file, collect and wrap the `main` function into a test case.
    """
    from _pytest.monkeypatch import MonkeyPatch
    from _pytest.python import Function

    path = Path(filepath)

    # Temporarily add parent directory to module search path.
    with MonkeyPatch.context() as m:
        m.syspath_prepend(path.parent)

        # Import file as Python module.
        mod = importlib.import_module(path.stem)
        fun = getattr(mod, entrypoint)

        # Wrap the entrypoint function into a pytest test case, and run it.
        test = Function.from_parent(request.node, name=entrypoint, callobj=fun)
        test.runtest()
        return test.reportinfo()


def pytest_notebook(request: pytest.FixtureRequest, filepath: t.Union[str, Path]):
    """
    From individual Jupyter Notebook file, collect cells as pytest
    test cases, and run them.

    Not using `NBRegressionFixture`, because it would manually need to be configured.
    """
    from _pytest._py.path import LocalPath
    from pytest_notebook.plugin import pytest_collect_file

    tests = pytest_collect_file(LocalPath(filepath), request.node)
    if not tests:
        raise ValueError(f"No tests collected from notebook: {filepath}")
    infos = []
    for test in tests.collect():
        test.runtest()
        infos.append(test.reportinfo())
    return infos
